///////////////////////////////////////////////////////////////////////////
//
//   Copyright 2010 Alberto González Palomo
//   Author: Alberto González Palomo - http://sentido-labs.com/
//
//   This file is part of Tutorial Toolkit, abbreviated as TTK.
//
//   TTK is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 3 of the License, or
//   (at your option) any later version.
//
//   TTK is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with TTK; if not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

package com.sentidolabs.ttk;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.*;
import javax.swing.text.html.HTMLDocument;
import java.util.Vector;
import java.util.Iterator;

/**
 * A call-out is a note that might be attached to one or more
 * GUI components in the underlaying application,
 * and is placed on a {@link Stencil} applied to the application's window.
 * </p><p>
 * This class provides the basic functionality but is not meant
 * to be instantiated, only used as base for the actual classes
 * that customize the appearance of the call-out by overriding
 * {@link #paintComponent(Graphics)}
 * </p>
 */
public abstract class CallOut extends JPanel
{
    protected Paint strokePaint;
    protected Stroke stroke;
    protected Paint background;
    protected Highlight highlight;
    protected int fontSize;
    protected String fontFamily;
    
    protected Content content;
    protected Point   contentLocation;
    protected Vector<Target> targets;
    protected Vector<Target> highlightTargets;
    protected Vector<Line2D> connectors;
    protected boolean targetsChanged;
    
    protected class Content extends JPanel
    {
        public JPanel controls;
        
        public Content()
        {
            controls = new JPanel();
            setLayout(new BorderLayout());
            add(controls, BorderLayout.SOUTH);
            setOpaque(false);
            controls.setOpaque(false);
            setVisible(true);
        }
        
        public void paintComponent(Graphics g)
        {
            Graphics2D g2 = (Graphics2D) g;
            g2.setPaint(background);
            g2.fillRect(0, 0, getWidth(), getHeight());
        }
    }
    
    protected class Target
    {
        public JComponent component;
        public Rectangle bounds;
        public Line2D connector;
        public Target(JComponent targetComponent)
        {
            component = targetComponent;
            bounds    = null;
            connector = null;
        }
    }
    
    protected CallOut()
    {
        content = new Content();
        targets          = new Vector<Target>();
        highlightTargets = new Vector<Target>();
        connectors       = new Vector<Line2D>();
        targetsChanged = false;
        
        super.add(content);
        setStroke(Color.BLACK, 2.0f);
        setBackground(Color.WHITE);
        setHighlight(new Highlight());
        setOpaque(false);
        setVisible(true);
        
        Font font = UIManager.getFont("Label.font");
        fontSize = font.getSize();
        fontFamily = font.getFamily();
    }
    
    /**
     * Set the stroke for border drawing.
     *
     * @param paint colour, texture, etc.
     * @param thickness floating-point number in pixels, e.g. 2.0f
     */
    public void setStroke(Paint paint, float thickness)
    {
        strokePaint = paint;
        stroke = new BasicStroke(thickness,
                                 BasicStroke.CAP_BUTT,
                                 BasicStroke.JOIN_BEVEL);
        repaint();
    }
    
    /**
     * Set the background paint.
     *
     * @param paint colour, texture, etc.
     */
    public void setBackground(Paint paint)
    {
        background = paint;
        repaint();
    }
    public void setBackground(Color colour)
    {
        setBackground((Paint) colour);
    }
    
    /**
     * Set the highlight to be used for hightlight targets.
     *
     * @param highlight the highlight style to be used
     */
    public void setHighlight(Highlight highlight)
    {
        this.highlight = highlight;
    }
    
    /**
     * Add a target for the call-out.
     *
     * @param target component that gets this call-out assigned to it
     */
    public void addTarget(JComponent target)
    {
        if (null == target) throw new RuntimeException("target can not be null");
        targets.add(new Target(target));
        targetsChanged = true;
        invalidate();
    }
    
    /**
     * Add a highlight target.
     *
     * @param target component that gets this highlight assigned to it
     */
    public void addHighlightTarget(JComponent target)
    {
        if (null == target) throw new RuntimeException("target can not be null");
        highlightTargets.add(new Target(target));
        targetsChanged = true;
        invalidate();
    }
    
    /**
     * Set the content of the call-out to the given HTML text.
     *
     * @param innerHTML new content as HTML
     * @return a new JTextPane with the given content
     */
    public Component setContent(String innerHTML)
    {
        JTextPane htmlView = new JTextPane();
        htmlView.setEditable(false);
        htmlView.setContentType("text/html");
        htmlView.setText(innerHTML);
        htmlView.setOpaque(false);
        ((HTMLDocument) htmlView.getDocument()).getStyleSheet()
            .addRule("body { "
                     + "font-family: " + fontFamily + "; "
                     + "font-size: "   + (fontSize * 1.25) + "pt; "
                     + "}");
        content.add(htmlView, BorderLayout.CENTER);
        
        return htmlView;
    }
    
    /**
     * Get the bounding box of the call-out's content.
     *
     * @return bounding rectangle
     */
    public Rectangle getContentBounds()
    {
        // We don't need to compute the absolute position because the
        // call-out occupies the whole stencil area.
        return content.getBounds();
    }
    
    public Component add(Component component)
    {
        content.controls.add(component);
        
        return component;
    }
    
    
    public Component add(String name, Component component)
    {
        return add(component);
    }
    
    public Component add(Component component, int index)
    {
        return add(component);
    }
    
    public void add(Component component, Object constraints)
    {
        add(component);
    }
    
    /**
     * Paint the call-out.
     * This is the method that must be overriden in subclasses for
     * painting the call-out. Do not override {@link #paint(Graphics)}
     *
     * When this function is called, the variable
     * <code>targetsChanged</code>
     * is <code>true</code> if there have been changes in the position
     * or size of the targets, so that any cached computations can be
     * updated.
     *
     * @param g graphics context where the call-out will be painted
     */
    public abstract void paintComponent(Graphics g);
    
    public void paint(Graphics g)
    {
        Graphics2D g2 = (Graphics2D) g;
        
        checkTargetChanges();
        repaint();// This is for the desktop pane, so that all internal frames are repainted.
        if (contentLocation != null) content.setLocation(contentLocation);
        
        super.paint(g);
        
        Iterator<Target> targetIterator = highlightTargets.iterator();
        while (targetIterator.hasNext()) {
            Rectangle highlighted = targetIterator.next().bounds;
            highlight.paintAround(highlighted, g2);
        }
        
        targetsChanged = false;
    }
    
    public void doLayout()
    {
        super.doLayout();
        checkTargetChanges();
        contentLocation = findPlaceFor(content.getSize());
        content.setLocation(contentLocation);
    }
    
    protected Line2D computeConnector(JComponent targetComponent)
    {
        return computeConnector(Stencil.getBoundsAbsolute(this, targetComponent));
    }
    protected Line2D computeConnector(Rectangle target)
    {
        int x1, y1, x2, y2;
        x1 = contentLocation.x + content.getWidth()  / 2;
        y1 = contentLocation.y + content.getHeight() / 2;
        int deltaX, deltaY;
        deltaX = target.x + target.width /2 - x1;
        deltaY = target.y + target.height/2 - y1;
        if (Math.abs(deltaX) > Math.abs(deltaY)) {
            if (target.x > x1) x2 = target.x;
            else               x2 = target.x + target.width;
            y2 = target.y + target.height / 2;
        }
        else {
            x2 = target.x + target.width / 2;
            if (target.y > y1) y2 = target.y;
            else               y2 = target.y + target.height;
        }
        
        return new Line2D.Double(x1, y1, x2, y2);
    }
    
    protected void checkTargetChanges()
    {
        Iterator<Target> targetIterator;
        
        targetIterator = targets.iterator();
        while (targetIterator.hasNext()) {
            Target target = targetIterator.next();
            Rectangle currentBounds = Stencil.getBoundsAbsolute(this, target.component);
            if (! currentBounds.equals(target.bounds)) {
                targetsChanged = true;
                target.bounds = currentBounds;
            }
        }
        
        targetIterator = highlightTargets.iterator();
        while (targetIterator.hasNext()) {
            Target target = targetIterator.next();
            Rectangle currentBounds = Stencil.getBoundsAbsolute(this, target.component);
            if (! currentBounds.equals(target.bounds)) {
                target.bounds = currentBounds;
            }
        }
        
        if (targetsChanged) {
            contentLocation = findPlaceFor(content.getSize());
            targetIterator = targets.iterator();
            while (targetIterator.hasNext()) {
                Target target = targetIterator.next();
                target.connector = computeConnector(target.bounds);
            }
        }
    }
    
    
    // Layout algorithm
    
    private class WeightedPoint
    {
        public int x, y;
        public double weightX, weightY;
        public WeightedPoint()
        {
            x = 0;
            y = 0;
            weightX = 0.0;
            weightY = 0.0;
        }
    }
    
    /**
     * Find a place for the given source that does not conflict with the targets.
     *
     * @param size dimensions of the source
     * @return new location for the source
     */
    protected Point findPlaceFor(Dimension size)
    {
        Point position = new Point(0, 0);
        Rectangle place = getBounds();
        WeightedPoint magnet;
        double x = 0.0, y = 0.0, totalWeigthX = 0.0, totalWeigthY = 0.0;
        Iterator<Target> targetIterator;
        
        targetIterator = targets.iterator();
        while (targetIterator.hasNext()) {
            Target target = targetIterator.next();
            magnet = awayFrom(Stencil.getBoundsAbsolute(this, target.component),
                              place,
                              size);
            x += magnet.x;
            y += magnet.y;
            totalWeigthX += magnet.weightX;
            totalWeigthY += magnet.weightY;
        }
        targetIterator = highlightTargets.iterator();
        while (targetIterator.hasNext()) {
            Target target = targetIterator.next();
            magnet = awayFrom(Stencil.getBoundsAbsolute(this, target.component),
                              place,
                              size);
            x += magnet.x;
            y += magnet.y;
            totalWeigthX += magnet.weightX;
            totalWeigthY += magnet.weightY;
        }
        
        if (totalWeigthX < 1.0) {
            x += place.x + place.width /2;
            totalWeigthX += 1.0;
        }
        if (totalWeigthY < 1.0) {
            y += place.y + place.height/2;
            totalWeigthY += 1.0;
        }
        x /= totalWeigthX;
        y /= totalWeigthY;
        
        position.x = (int) x - size.width /2;
        position.y = (int) y - size.height/2;
        
        int margin = 8;
        place.x += margin;
        place.y += margin;
        place.width  -= 2*margin;
        place.height -= 2*margin;
        
        if (position.x < place.x) {
            position.x = place.x;
        }
        else if (position.x + size.width > place.x + place.width) {
            position.x = place.x + place.width - size.width;
        }
        if (position.y < place.y) {
            position.y = place.y;
        }
        else if (position.y + size.height > place.y + place.height) {
            position.y = place.y + place.height - size.height;
        }
        
        return position;
    }
    
    private WeightedPoint awayFrom(Rectangle target, Rectangle place, Dimension size)
    {
        WeightedPoint magnet = new WeightedPoint();
        double targetCenterX, targetCenterY, centerX, centerY;
        centerX = place.getCenterX();
        centerY = place.getCenterY();
        targetCenterX = target.getCenterX();
        targetCenterY = target.getCenterY();
        int separationX = size.width, separationY = size.height;
        if (Math  .abs(targetCenterX - centerX) / place.getWidth()
            > Math.abs(targetCenterY - centerY) / place.getHeight()) {
            if (targetCenterX > centerX) {
                magnet.x = target.x - separationX - size.width/2;
            }
            else {
                magnet.x = target.x + target.width + separationX + size.width/2;
            }
            magnet.weightX = 1.0;
            magnet.weightY = 0.0;
        }
        else {
            if (targetCenterY > centerY) {
                magnet.y = target.y - separationY - size.height/2;
            }
            else {
                magnet.y = target.y + target.height + separationY + size.height/2;
            }
            magnet.weightX = 0.0;
            magnet.weightY = 1.0;
        }
        
        return magnet;
    }
    
}
