///////////////////////////////////////////////////////////////////////////
//
//   Copyright 2010 Alberto González Palomo
//   Author: Alberto González Palomo - http://sentido-labs.com/
//
//   This file is part of Tutorial Toolkit, abbreviated as TTK.
//
//   TTK is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 3 of the License, or
//   (at your option) any later version.
//
//   TTK is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with TTK; if not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

package com.sentidolabs.ttk;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;

/**
 * Semi-transparent overlay with holes and call-outs for certain UI components.
 * <p>
 * The underlaying application is only accessible through
 * the holes in the Stencil made with
 * {@link #addHoleFor(JComponent)}
 * </p><p>
 * Call-outs (notes with the feedback) are placed on the stencil with
 * {@link #add(Component)}
 * </p><p>
 * When making holes or adding covers in the stencil, a layer name can be
 * specified. Otherwise those operations are done in the default layer.
 * </p>
 */
public class Stencil extends JPanel
{
    static FocusTraversalDisablePolicy focusTraversalDisablePolicy;
    
    private Paint foilColour;
    private Area foilShape, foilDrawnShape;
    private boolean defaultCovered;
    private Component defaultGlassPane;
    private FocusTraversalPolicy defaultFocusTraversalPolicy;
    private Vector<CallOut> callOuts;
    private class Layer
    {
        private Vector<JComponent> holes;
        private Vector<JComponent> covers;
        public Layer()
        {
            holes  = new Vector<JComponent>();
            covers = new Vector<JComponent>();
        }
        public void addHoleFor(JComponent target)
        {
            if (null == target) throw new RuntimeException("target can not be null");
            holes.add(target);
        }
        public void addCoverFor(JComponent target)
        {
            if (null == target) throw new RuntimeException("target can not be null");
            covers.add(target);
        }
        public void reset()
        {
            holes .removeAllElements();
            covers.removeAllElements();
        }
    }
    private Hashtable<String, Layer> layers;
    private int opacity, red, green, blue;
    private boolean antialias = true;
    private JFrame parentWindow;
    private EventSilencer eventSilencer;
    
    /**
     * Constructs a new stencil overlaid on the given window.
     *
     * @param window the frame whose glass pane will be replaced by the stencil.
     *        The original glass pane is restored when calling {@link #detach()}
     */
    public Stencil(JFrame window)
    {
        this(window, true);
    }
    
    /**
     * Constructs a new stencil overlaid on the given window.
     * If it is set to not cover the whole window by default, only those components for which {@link #addCoverFor(JComponent)} is called will be covered.
     *
     * @param window the frame whose glass pane will be replaced by the stencil.
     * @param defaultCovered whether the stencil will cover the whole window by default or not
     */
    public Stencil(JFrame window, boolean defaultCovered)
    {
        foilShape      = new Area();
        foilDrawnShape = new Area();
        this.defaultCovered = defaultCovered;
        
        setColour(60, 60, 255);
        setOpacity(127);
        
        setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
        setOpaque(false);
        
        layers   = new Hashtable<String, Layer>();
        callOuts = new Vector<CallOut>();
        
        parentWindow = window;
        defaultGlassPane = window.getGlassPane();
        window.setGlassPane(this);
        
        eventSilencer = new EventSilencer();
        addMouseListener(eventSilencer);
        addMouseMotionListener(eventSilencer);
        addMouseWheelListener(eventSilencer);
        
        defaultFocusTraversalPolicy = window.getFocusTraversalPolicy();
        window.setFocusTraversalPolicy(new FocusTraversalDisablePolicy());
        
        setVisible(true);
        parentWindow.validate();
    }
    
    /**
     * Set the colour for the stencil overlay.
     *
     * @param red from 0 to 255
     * @param green from 0 to 255
     * @param blue from 0 to 255
     */
    public void setColour(int red, int green, int blue)
    {
        this.red   = red;
        this.green = green;
        this.blue  = blue;
        foilColour = new Color(red, green, blue, opacity);
        invalidate();
    }
    
    /**
     * Set the paint for the stencil overlay.
     * This allows painting the stencil with gradients, textures and other
     * effects instead of just a flat colour.
     * The paint object must already contain any opacity information:
     * calling {@link #setOpacity(int)} does not have any effect in this case
     * unless {@code paint} is an instance of {@link Color}.
     *
     * @see java.awt.Color
     * @see java.awt.SystemColor
     * @see java.awt.LinearGradientPaint
     * @see java.awt.MultipleGradientPaint
     * @see java.awt.RadialGradientPaint
     * @see java.awt.TexturePaint
     *
     * @param paint generic paint object that can be a colour, a gradient, etc.
     */
    public void setColour(Paint paint)
    {
        foilColour = paint;
        if (foilColour instanceof Color) {
            red     = ((Color) paint).getRed();
            green   = ((Color) paint).getGreen();
            blue    = ((Color) paint).getBlue();
            opacity = ((Color) paint).getAlpha();
        }
        invalidate();
    }
    
    /**
     * Set the opacity of the stencil overlay.
     * This call has no effect if the stencil colour is set with
     * {@link #setColour(Paint)} and the given paint is not
     * an instance of {@link Color}.
     *
     * @param opacity from 0 (fully transparent) to 255 (fully opaque)
     */
    public void setOpacity(int opacity)
    {
        if (!(foilColour instanceof Color)) return;
        
        this.opacity = opacity;
        foilColour = new Color(red, green, blue, opacity);
        invalidate();
    }
    
    /**
     * Detach the stencil from its window, restoring the window's original
     * glass pane, focus traversal policy, etc.
     *
     */
    public void detach()
    {
        if (parentWindow != null) {
            parentWindow.setFocusTraversalPolicy(defaultFocusTraversalPolicy);
            parentWindow.setGlassPane(defaultGlassPane);
            parentWindow.validate();
            parentWindow = null;
        }
    }
    
    /**
     * Punch a hole in the stencil for the given GUI component.
     * The is no limit for the number of holes in a given stencil.
     *
     * @param target component that will be accesible through the hole
     */
    public void addHoleFor(JComponent target)
    {
        addHoleFor(target, null);
    }
    /**
     * Punch a hole in the stencil for the given GUI component in the named layer.
     * The is no limit for the number of holes or layers in a given stencil.
     *
     * @param target component that will be accesible through the hole
     * @param layerName name of the stencil layer
     */
    public void addHoleFor(JComponent target, String layerName)
    {
        getLayer(layerName).addHoleFor(target);
        invalidate();
    }
    
    /**
     * Add a cover for the given target.
     * A cover is the inverse of a hole.
     * The is no limit for the number of covers in a given stencil.
     *
     * @param target component to be covered by the stencil
     */
    public void addCoverFor(JComponent target)
    {
        addCoverFor(target, null);
    }
    
    /**
     * Add a cover for the given target in the named layer.
     * A cover is the inverse of a hole.
     * The is no limit for the number of covers or layers in a given stencil.
     *
     * @param target component to be covered by the stencil
     * @param layerName name of the stencil layer
     */
    public void addCoverFor(JComponent target, String layerName)
    {
        getLayer(layerName).addCoverFor(target);
        invalidate();
    }
    
    /**
     * Remove all call-outs from the stencil.
     *
     */
    public void removeAll()
    {
        super.removeAll();
        callOuts.removeAllElements();
    }
    
    /**
     * Remove all holes and covers from the stencil.
     *
     */
    public void reset()
    {
        reset(null);
    }
    /**
     * Remove all holes and covers from the stencil in the named layer.
     *
     */
    public void reset(String layerName)
    {
        getLayer(layerName).reset();
        invalidate();
        repaint();
    }
    
    /**
     * Add a {@link CallOut} component to the stencil.
     *
     * @param component component to add, ignored if it is not a call-out
     * @return component added
     */
    public Component add(Component component)
    {
        if (component instanceof CallOut) {
            super.add(component);
            callOuts.add((CallOut) component);
        }
        else {
            System.err.println("Warning: the TTK Stencil only accepts CallOut components");
        }
        
        return component;
    }
    
    /**
     * This is the same as {@link #add(Component)}
     *
     * @param name component's name, ignored
     * @param component component to add, ignored if it is not a call-out
     * @return component added
     */
    public Component add(String name, Component component)
    {
        return add(component);
    }
    
    /**
     * This is the same as {@link #add(Component)}
     *
     * @param component component to add, ignored if it is not a call-out
     * @param index component's index, ignored
     * @return component added
     */
    public Component add(Component component, int index)
    {
        return add(component);
    }
    
    /**
     * This is the same as {@link #add(Component)}
     *
     * @param component component to add, ignored if it is not a call-out
     * @param constraints layout constraints, ignored
     */
    public void add(Component component, Object constraints)
    {
        add(component);
    }
    
    /**
     * Compute the bounds of a component with respect to a container.
     *
     * @param container the reference container
     * @param component the element whose bounds we want to compute
     * @return bounding rectangle in coordinates relative to the container
     */
    public static Rectangle getBoundsAbsolute(Container container, JComponent component)
    {
        if (container == null) System.err.println("container is null");
        if (component == null) System.err.println("component is null");
        Rectangle rectangle = component.getBounds();
        int containerX = container.getX(), containerY = container.getY();
        Component parent;
        parent = container;
        while ((parent = parent.getParent()) != null) {
            containerX += parent.getX();
            containerY += parent.getY();
        }
        parent = component;
        while ((parent = parent.getParent()) != null) {
            rectangle.x += parent.getX();
            rectangle.y += parent.getY();
        }
        
        rectangle.x -= containerX;
        rectangle.y -= containerY;
        
        return rectangle;
    }
    
    private Layer getLayer(String layerName)
    {
        Layer layer;
        if (null == layerName) layerName = "";
        if (layers.containsKey(layerName)) {
            layer = layers.get(layerName);
        }
        else {
            layer = new Layer();
            layers.put(layerName, layer);
        }
        
        return layer;
    }
    
    private Area cutoutFor(JComponent component)
    {
        return cutoutFor(component, 3);
    }
    private Area cutoutFor(JComponent component, int margin)
    {
        Rectangle rectangle = getBoundsAbsolute(this, component);
        rectangle.x -= margin;
        rectangle.y -= margin;
        rectangle.width  += margin + margin;
        rectangle.height += margin + margin;
        return new Area(rectangle);
    }
    
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D) g;
        if (antialias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                RenderingHints.VALUE_ANTIALIAS_ON);
        }
        
        compose();
        
        Shape originalClip = g2.getClip();
        g2.clip(foilDrawnShape);
        g2.setPaint(foilColour);
        g2.fillRect(0, 0, getWidth(), getHeight());
        g2.setClip(originalClip);
    }
    
    private void compose()
    {
        Rectangle bounds;
        bounds = getBounds();
        bounds.x = 0;
        bounds.y = 0;
        foilShape     .reset();
        foilDrawnShape.reset();
        if (defaultCovered) {
            foilShape     .add(new Area(bounds));
            foilDrawnShape.add(new Area(bounds));
        }
        
        Iterator<Layer> layerIterator = layers.values().iterator();
        while (layerIterator.hasNext()) {
            Layer layer = layerIterator.next();
            Iterator<JComponent> holeIterator = layer.holes.iterator();
            while (holeIterator.hasNext()) {
                JComponent hole = holeIterator.next();
                foilShape     .subtract(cutoutFor(hole, 0));
                foilDrawnShape.subtract(cutoutFor(hole));
            }
            Iterator<JComponent> coverIterator = layer.covers.iterator();
            while (coverIterator.hasNext()) {
                JComponent cover = coverIterator.next();
                foilShape     .add(cutoutFor(cover, 0));
                foilDrawnShape.add(cutoutFor(cover, 0));
            }
        }
        
        Iterator<CallOut> callOutIterator = callOuts.iterator();
        while (callOutIterator.hasNext()) {
            CallOut callOut = callOutIterator.next();
            foilShape.add(new Area(callOut.getContentBounds()));
        }
    }
    
    // Event handling
    private class FocusTraversalDisablePolicy
        extends FocusTraversalPolicy
    {
        private Vector<JComponent> holes;
        
        private FocusTraversalDisablePolicy() {}
        
        public FocusTraversalDisablePolicy(Vector<JComponent> holes)
        {
            this.holes = holes;
        }
        
        public Component getComponentAfter(Container root, Component component)
        {
            if (holes == null || holes.size() == 0) return null;
            int i;
            for (i = 0; i < holes.size(); ++i) {
                if (holes.elementAt(i) == component) {
                    ++i;
                    if (i < holes.size()) component = holes.elementAt(i);
                    else                  component = holes.firstElement();
                    break;
                }
            }
            
            return component;
        }
        public Component getComponentBefore(Container root, Component component)
        {
            if (holes == null || holes.size() == 0) return null;
            int i;
            for (i = 0; i < holes.size(); ++i) {
                if (holes.elementAt(i) == component) {
                    --i;
                    if (i >= 0) component = holes.elementAt(i);
                    else        component = holes.lastElement();
                    break;
                }
            }
            
            return component;
        }
        public Component getLastComponent(Container root)
        {
            if (holes == null || holes.size() == 0) return null;
            
            return holes.lastElement();
        }
        public Component getFirstComponent(Container root)
        {
            if (holes == null || holes.size() == 0) return null;
            
            return holes.firstElement();
        }
        public Component getDefaultComponent(Container root)
        {
            return getFirstComponent(root);
        }
    }
    
    public boolean contains(int x, int y)
    {
        return foilShape.contains(new Point(x, y));
    }
    
    public boolean isOptimizedDrawingEnabled()
    {
        return false;
    }
    
    private class EventSilencer implements MouseInputListener, MouseWheelListener
    {
        // MouseInputListener = MouseListener + MouseMotionLister
        // MouseListener
        public void mouseClicked(MouseEvent e) {}
        
        public void mouseEntered(MouseEvent e) {}
        
        public void mouseExited(MouseEvent e) {}
        
        public void mousePressed(MouseEvent e) {}
        
        public void mouseReleased(MouseEvent e) {}
        
        // MouseMotionListener
        public void mouseDragged(MouseEvent e) {}
        
        public void mouseMoved(MouseEvent e) {}
        
        // MouseWheelListener
        public void mouseWheelMoved(MouseWheelEvent e) {}
        
    }
    
}
